# coding: utf-8
import hues

from django.core.management.base import BaseCommand

from elasticsearch import exceptions
from elasticsearch_dsl.connections import connections
from elasticsearch_flex.indexes import registered_indices


class Command(BaseCommand):
    help = 'Sync search indices, templates, and scripts.'

    def handle(self, *args, **options):
        indices = registered_indices()
        connection = connections.get_connection()
        hues.info('Using connection', connection)
        if len(indices):
            hues.info('Discovered', len(indices), 'Indexes')
        else:
            hues.warn('No search index found')
        for i, index in enumerate(indices, 1):
            hues.info('==> Initializing', index.__name__)

            index_name = index._doc_type.index
            try:
                connection.indices.close(index_name)
            except exceptions.NotFoundError:
                pass
            index.init()
            connection.indices.open(index_name)

            hues.success('--> Done {0}/{1}'.format(i, len(indices)))



import six
from six.moves import reduce


def rgetattr(obj, attr, allow_null=True):
    attrs = attr.split('__')

    def _get_attr(o, a):
        try:
            return getattr(o, a)
        except AttributeError as e:
            if not allow_null:
                raise e

    return reduce(_get_attr, [obj] + attrs)


def memoized_property(fget):
    """
    Return a property attribute for new-style classes that only calls its getter
    on the first access. The result is stored and on subsequent accesses is
    returned, preventing the need to call the getter any more.

    Example:
    Following classes C, and D are equivalent.
    >>> class C(object):
    ...   @memoized_property
    ...   def name(self):
    ...     return expensive_function_call()
    >>> class D(object):
    ...   @property
    ...   def name(self):
    ...     if not hasattr(self, '_name'):
    ...       self._name = expensive_function_call()
    ...     return self._name
    """
    attr_name = '_{0}'.format(fget.__name__)

    @six.wraps(fget)
    def fget_memoized(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fget(self))
        return getattr(self, attr_name)

    return property(fget_memoized)



import elasticsearch
import json
import logging
import os
import re
import six

from . import connections

logger = logging.getLogger(__name__)


@six.python_2_unicode_compatible
class SearchTemplate(object):
    '''Partial Search Templates

    Wrapper for lower level Search Templates stored as mustache templates in
    elasticsearch cluster.

    A template is a `json` file containing the elasticsearch query with
    placeholders for parameters passed during the queries. It is possible to
    also optionally specify a `script` file name, which is inlined before
    it is registered in elasticsearch.

    To specify the `script`, use the following substitution:

            "field": "#[filename]",
    '''

    def __init__(self, name, filepath):
        self.name = name
        self.template = self._prerender(filepath)

    def register(self):
        c = connections.get_connection()
        try:
            c.put_template(self.name, self.template)
            logger.info('Registered template %s', self)
        except elasticsearch.TransportError:
            logger.exception('Error while attempting to PUT template <%s>', self.name)
            raise

    def unregister(self):
        c = connections.get_connection()
        c.delete_template(self.name)

    def _prerender(self, filepath):
        file_abspath = os.path.abspath(filepath)
        file_dir = os.path.dirname(file_abspath)
        with open(filepath, 'r') as fp:
            content = json.load(fp)

        # Recurse and look for a #[filename] pattern and load that file.
        def interpolate(dat):
            if type(dat) is dict:
                return {k: interpolate(v) for k, v in dat.items()}
            elif type(dat) is list:
                return [interpolate(x) for x in dat]
            elif type(dat) is six.text_type:
                matches = re.match(r'#\[(.+)\]', dat)
                if matches is not None:
                    script_name = matches.group(1)
                    script_file = os.path.join(file_dir, '{}.java'.format(script_name))
                    with open(script_file, 'r') as fp:
                        script = fp.read()
                        # XXX: Very cheap Java code "minification".
                        # It looks for spaces and tab marks in the beginning
                        # and end of the string, and replaces them with a single
                        # space.
                        # This is not robust and will fail for multi-line
                        # strings which want to preserve the spaces, however
                        # in our use case, we probably won't need that.
                        # In any case, this can use a rewrite.
                        return re.sub(r'^[ \t]+|\s+$', ' ', script, flags=re.M)
            return dat

        prerendered = interpolate(content)
        body = json.dumps(prerendered)

        # We need to fix the above (valid) json to cater for being a template
        # Remove the quotes around the json props.
        body = (body
                .replace('"{{#toJson}}', "{{#toJson}}")
                .replace('{{/toJson}}"', '{{/toJson}}'))
        return json.dumps(dict(template=body))

    def __repr__(self):
        return '<{0} name={1}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        return repr(self)



# coding: utf-8
import hues
from tqdm import tqdm

from django.core.management.base import BaseCommand

from elasticsearch_flex.indexes import registered_indices


class Command(BaseCommand):
    help = 'Setup Search Index'

    def handle(self, *args, **options):
        indices = registered_indices()
        if len(indices):
            hues.info('Discovered', len(indices), 'Indexes')
        else:
            hues.warn('No search indexes found')
        for i, index in enumerate(indices, 1):
            hues.info('==> Indexing objects from', index.model, 'in', index)
            for obj in tqdm(index.queryset):
                doc = index.init_using_pk(obj.pk)
                doc.prepare()
                doc.save()
            hues.success('--> {0}/{1} indexed.'.format(i, len(indices)))



import logging

from django.conf import settings
from elasticsearch_dsl.connections import connections

__version__ = '0.2.2'

logger = logging.getLogger('elasticsearch_flex')

default_app_config = 'elasticsearch_flex.apps.ElasticsearchFlexConfig'

flexconfig = dict()

if hasattr(settings, 'ELASTICSEARCH_FLEX'):
    flexconfig.update(settings.ELASTICSEARCH_FLEX)

__all__ = ('connections',)



import os
import inspect
import six

from collections import namedtuple
from six.moves import filter

from elasticsearch_dsl.document import DocType, DocTypeMeta

from .fields import FlexField
from .utils import rgetattr
from .query import DocAccessors
from .search_templates import SearchTemplate

_MODEL_INDEX_MAPPING = {}

FieldAttrMap = namedtuple('FieldAttrMap', ('name', 'attr'))


class IndexableMeta(type):
    '''Defines an introspection type for Indexable Models.'''
    def __new__(cls, name, bases, namespace):
        fields = []
        # We introspect and collect the names of all the Fields in Index
        # to facilitate differentiating field attribute from non-field attribute
        # later.
        for k, x in namespace.items():
            if isinstance(x, FlexField):
                f = x._model_attr if x._model_attr is not None else k
                fields.append(FieldAttrMap(k, f))
        # Append field-name information from all the Base classes.
        ix_predicate = lambda x: issubclass(x, DocType) and hasattr(x, '_fields')
        for base_index in filter(ix_predicate, bases):
            fields.extend(base_index._fields)
        # Insert fields into this class's namespace.
        namespace['_fields'] = fields

        # Make a reference of Meta in the object namespace (would not be GCd)
        if 'Meta' in namespace:
            namespace['_meta'] = namespace['Meta']

        # TODO: Assert sanity of extension class
        # -- Implement checking get_queryset and get_model implementations.
        klass = super(IndexableMeta, cls).__new__(cls, name, bases, namespace)
        # Register this doctype in indexed models mapping.
        if 'model' in namespace:
            _MODEL_INDEX_MAPPING[namespace['model']] = klass
        klass.docs = DocAccessors(klass)
        return klass


class IndexableDocTypeMeta(IndexableMeta, DocTypeMeta):
    pass


@six.python_2_unicode_compatible
@six.add_metaclass(IndexableDocTypeMeta)
class IndexedModel(DocType):
    '''Base class for declaring Index for a Model.

    This class is an extension of Elasticsearch-DSL's DocType. For the fields,
    as a convention, it is assumed that the model has the attribute of the
    same name. This default behaviour can be controlled by implementing a
    `prepare_<FieldName>` method, accepting the object instance.
    '''
    def get_queryset(self):
        return self.queryset

    def get_object(self):
        qs = self.get_queryset()
        return qs.get(pk=self._id)

    @property
    def object(self):
        return self.get_object()

    def prepare(self):
        obj = self.get_object()
        for field in self._fields:
            try:
                prep_method = getattr(self, 'prepare_{0}'.format(field.attr))
                setattr(self, field.name, prep_method(obj))
            except AttributeError:
                val_from_obj_attr = rgetattr(obj, field.attr)
                setattr(self, field.name, val_from_obj_attr)

    @classmethod
    def init_using_pk(cls, pk):
        obj = cls(_id=pk)
        return obj

    @classmethod
    def init(cls, index=None, using=None):
        ix_name = cls._doc_type.index
        module_dir = os.path.dirname(inspect.getabsfile(cls))
        tpl_dir = os.path.join(module_dir, 'search_templates')

        # Init mapping
        cls._doc_type.init(index, using)

        # Discover and register the Templates
        if hasattr(cls, '_meta'):
            for tpl in getattr(cls._meta, 'query_templates', []):
                identifier = '.'.join([ix_name, tpl])
                template_file = os.path.join(tpl_dir, '{}.json'.format(tpl))
                template = SearchTemplate(identifier, template_file)
                template.register()

    def __repr__(self):
        return '<{0} index={1}>'.format(self.__class__.__name__, self._index)

    def __str__(self):
        return repr(self)


def get_index_for_model(model):
    try:
        return _MODEL_INDEX_MAPPING[model]
    except KeyError:
        raise KeyError('Model {0} has no associated index'.format(model))


def get_model_for_index(index):
    try:
        rev_map = (m for m, i in _MODEL_INDEX_MAPPING.items() if i is index)
        return next(rev_map, None)
    except StopIteration:
        raise KeyError('Index {0} has no associated model'.format(index))


def registered_indices():
    return _MODEL_INDEX_MAPPING.values()

__all__ = (
    'IndexedModel',
    'get_index_for_model',
    'get_model_for_index',
    'registered_indices',
)



import logging

from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver

from . import tasks
from .indexes import get_index_for_model

logger = logging.getLogger('elasticsearch_flex.signals')


@receiver(post_save)
def sync_elasticsearch_on_model_update_or_create(sender, instance, created, **kwargs):
    try:
        ix = get_index_for_model(sender)
    except KeyError:
        pass
    else:
        tasks.update_indexed_document.delay(ix, created, instance.pk)


@receiver(post_delete)
def sync_elasticsearch_on_model_delete(sender, instance, **kwargs):
    try:
        ix = get_index_for_model(sender)
    except KeyError:
        pass
    else:
        tasks.delete_indexed_document.delay(ix, instance.pk)


__all__ = (
    'sync_elasticsearch_on_model_update_or_create',
    'sync_elasticsearch_on_model_delete',
)
